#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <iostream>
#include <string>
using namespace std;
#include "strtoken.h"
#include "cycle.h"

int reg(int n);
void set_reg(int n, int value);
int mem_check(int addr);
int get_mem(int addr, int &q);
int set_mem(int addr, int data);
void showpc(FILE *out, int pc, bool color);


string cmd[] = {  "NOP", "LWI", "SWI", "LI", "LUI",
"LW", "SW", "ADDI", "ALU1", "ALU2", "BNE", "JALR"};

string cmd_alu1[] = { "ADD", "SUB" };
string cmd_alu2[] = { "AND", "OR", "XOR", "NOR" };

const int  NOP = 0, LWI = 1, SWI = 2,  LI = 3, LUI = 4,
LW = 5, SW=6, ADDI = 7, ALU1 = 8, ALU2 = 9, BNE = 10, JALR = 11;

const int HALT = 0;



string opcode_memnomic(int opcode, int opx)
{
	char str[64];
	if (opcode==ALU1) return cmd_alu1[opx];
	if (opcode==ALU2) return cmd_alu2[opx];
	if (opcode<12) return cmd[opcode];
	sprintf_s(str,63,"NOP: %0X",opcode);
	return str;
}
void cycle::show(FILE *out, int ncycle)
{
	fprintf(out,"<tr><td bgcolor=FFFFCC>%d<td>%02X",ncycle+1,pc);
	fprintf(out,"<td>%s<td>%d<td>%04X<td>",
		  opcmd.c_str(),rd,qd);
	if (ma) fprintf(out,"%04X\n",addr);
	else fprintf(out,"<br>\n");
}

void cycle::set_addr(int maddr)
{
	addr = maddr;
	ma = true;
}


int step_cpu(int inst, int &nextpc, cycle &step)
{
	int pc;
	int immed, immse, maddr;
	int opx, imm5;
	int opcode, rd=0, ra=0, rb=0;
	int qa, qb, qd=0;
	bool ma;

	opcode = inst>>11;
	rd = (inst>>8)&0x7;
	ra = (inst>>5)&0x7;
	rb = (inst>>2)&0x7;
	immed = inst&0xFF;
	if (immed>>7) immse = 0xFFFFFF00| immed;
	else immse = immed;
	imm5 = inst&0x1F;
	if (imm5>>4) imm5 = 0xFFFFFFE0 | imm5;
	qa = reg(ra);
	qb = reg(rb);
	maddr = immed;
	pc = nextpc;
	nextpc++;
	opx = immed&0x3;
	ma = false;
	step.opcmd = opcode_memnomic(opcode,opx);
	switch (opcode) {
		case LI:
			qd = immse;
			break;
		case LUI:
			ra = rd;
			qa = reg(ra);
			qd = (immed<<8)|(qa&0xFF);
			break;
		case SWI: // SW
			rb = rd;
			rd = 0;
			qb = reg(rb);
			ma = true;
			maddr = immed;
			qd = qb;
			set_mem(maddr,qb);
			break;
		case SW:
			rb = rd;
			rd = 0;
			qb = reg(rb);
			ma = true;
			maddr = qa+imm5;
			qd = qb;
			set_mem(maddr,qb);
			break;
		case LWI:
			ma = true;
			maddr = immed;
			get_mem(maddr,qd);
			break;
		case LW:
			ma = true;
			maddr = reg(ra)+imm5;
			get_mem(maddr,qd);
			break;
		case ADDI:
			qd = qa + imm5;
			break;
		case ALU1:
			switch (opx) {
				case 0:
					qd = qa+qb; // add
					break;
				case 1:
					qd = qa-qb; // subtract
					break;
			}
			break;
		case ALU2:
			switch (opx) {
				case 0:
					qd = qa & qb;
					break;
				case 1:
					qd = qa | qb;
					break;
				case 2:
					qd = qa ^ qb;
					break;
				case 3:
					qd = ~(qa|qb);
					break;
			}
			break;
		case BNE:
			rb = rd;
			rd = 0;
			qd = reg(rb)!=qa;
			if (qd) nextpc = nextpc + imm5;
			break;
		case JALR:
			qd = nextpc;
			nextpc = qa;
			break;
		default:
			rd = 0;
			break;
	}
	qd &= 0xFFFF;
	bool we = opcode!=SW && opcode!=SWI;
	if (rd && we) set_reg(rd,qd);
	step.set(pc,rd,qd);
	if (ma) step.set_addr(maddr);
	if (inst==HALT) {
		step.opcmd="HALT";
		return 1;
	}
	return 0;
}

void IFID::set(int inst, int nextpc)
{
	int opcode;
	int rdx;
	int immed, immse, imm5;
	instr = inst;
	opcode = inst>>11;
	pc = nextpc;
	op = opcode;
	opx = inst&0x3;
	rdx = (inst>>8)&0x7;
	ra = (inst>>5)&0x7;
	rb = (inst>>2)&0x7;
	immed = inst&0xFF;
	if (immed>>7) immse = 0xFFFFFF00| immed;
	else immse = immed;
	imm5 = inst&0x1F;
	if (imm5>>4) imm5 = 0xFFFFFFE0 | imm5;

	immed = inst&0xFF;
	if (immed>>7) immse = 0xFFFFFF00| immed;
	else immse = immed;
	imm5 = inst&0x1F;
	if (imm5>>4) imm5 = 0xFFFFFFE0 | imm5;
	bool sw = opcode==SW || opcode==SWI;
	bool flag = sw || opcode==BNE;
	ra = opcode==LUI? rdx: ra;
	rd = flag? 0: rdx;
	rb = (opcode==ALU1 || opcode==ALU2)? rb:
		      (flag? rdx: 0);
}



void IFID::show(FILE *out)
{
	string cmd = opcode_memnomic(op,opx);
	if (active) {
		showpc(out,pc,color);
		fprintf(out,"<br>%s",halt? "halt":cmd.c_str());
	}
	else fprintf(out,"<TD><BR>");
}

void IFID::bubble()
{
	active = false;
	halt = false;
	instr = 0;
	op = 0;
	rd = 0;
}


void IDEX::bubble()
{
	active = false;
	halt = false;
	instr = 0;
	rd = 0;
}

void IDEX::set(IFID &old)
{
	active = old.active;
	color = old.color;
	instr = old.instr;
	halt = old.halt;
	pc = old.pc;
	rd = old.rd;
	ra = old.ra;
	rb = old.rb;
	qa = reg(ra);
	qb = reg(rb);
}

void IDEX::show(FILE *out)
{
	if (active) {
		showpc(out,pc,color);
		fprintf(out,"<br>%d %d %d\n",rd,ra,rb);
	}
	else fprintf(out,"<td><br>");
}

void EXWB::set(IDEX &old)
{
	active = old.active;
	color = old.color;
	halt = old.halt;
	pc = old.pc;
	instr = old.instr;
	rd = old.rd;
	ra = old.ra;
	rb = old.rb;
	qa = reg(ra);
	qb = reg(rb);
}

void EXWB::forward(IDEX &IDEX_old, EXWB &EXWB_old, FILE *out, int ncycle)
{
	if (IDEX_old.ra == EXWB_old.rd && IDEX_old.ra !=0) {
		qa = EXWB_old.qd;
		fprintf(stderr,"(cycle %d) qa forwarded: %04x\n",ncycle+1,qa);
		fprintf(out,"<tr align=center><td><br>");
		fprintf(out,"<td bgcolor=FFFF00 colspan=5>qa forwarded: %04x\n",qa);
	}
	if (IDEX_old.rb == EXWB_old.rd && IDEX_old.rb !=0) {
		qb = EXWB_old.qd;
		fprintf(stderr,"(cycle %d) qb forwarded: %04x\n",ncycle+1,qb);
		fprintf(out,"<tr align=center><td><br>");
		fprintf(out,"<td bgcolor=FFFF00 colspan=5>qb forwarded: %04x\n",qb);
	}
}

int EXWB::execute(FILE *out, IFID &IFID_new, IDEX &IDEX_new)
{
	int opcode, opx;
	int immed, immse, imm5;
	int nextpc = pc+1;
	opcode = instr>>11;
	opx = instr&0x3;
	immed = instr&0xFF;
	if (immed>>7) immse = 0xFFFFFF00| immed;
	else immse = immed;
	imm5 = instr&0x1F;
	if (imm5>>4) imm5 = 0xFFFFFFE0 | imm5;
	maddr = immed;
	if (opcode==LW || opcode==SW) maddr = qa + imm5;
	jump = -1;
	ma = false;
	switch (opcode) {
		case LI:
			qd = immse;
			break;
		case LUI:
			qd = (immed<<8) | (qa&0xff);
			break;
		case SWI:
		case SW:
			ma = true;
			qd = qb;
			set_mem(maddr,qb);
			break;
		case LWI:
		case LW:
			ma = true;
			get_mem(maddr,qd);
			break;
		case ADDI:
			qd = qa + imm5;
			break;
		case ALU1:
			switch (opx) {
				case 0:
					qd = qa+qb; // add
					break;
				case 1:
					qd = qa-qb; // subtract
					break;
			}
			break;
		case ALU2:
			switch (opx) {
				case 0:
					qd = qa & qb;
					break;
				case 1:
					qd = qa | qb;
					break;
				case 2:
					qd = qa ^ qb;
					break;
				case 3:
					qd = ~(qa|qb);
					break;
			}
			break;

		case BNE:
			qd = qa!=qb;
			if (qd) {
				jump = nextpc + imm5; // take branch
				IFID_new.bubble();
				IDEX_new.bubble();
				//bubble();
				}
			if (qd && out) {
				fprintf(out,"<tr align=center><td><br>");
				fprintf(out,"<td bgcolor=8080FF colspan=5>branch %04x\n",jump);
			}
			break;
		case JALR:
			qd = nextpc;
			jump = qa;
			// insert bubbles
			IFID_new.bubble();
			IDEX_new.bubble();
			if (out) {
			fprintf(out,"<tr align=center><td><br>");
			fprintf(out,"<td bgcolor=FFC0C0 colspan=5> nextpc %04x jump %04x\n",qd,qa);
			}
			break;

	}
	qd &= 0xFFFF;
	return jump;
}


void EXWB::showmem(FILE *out)
{
	int op = instr>>11;
	if (op==LW || op==LWI) fprintf(out,"qd = %04x from mem[0x%x]\n",qd,maddr);
	if (op==SW || op==SWI) fprintf(out,"mem[0x%x] = %04x\n",maddr,qd);
}

void EXWB::show(FILE *out)
{
	int opcode = instr>>13;
	bool flag = (opcode==LW || opcode==SW);
	if (active) {
		showpc(out,pc,color);
		fprintf(out,"<br>%04x",(flag? maddr: qd));
	}
	else fprintf(out,"<td><br>");
}


void WBXX::set(EXWB &old)
{
int opcode;
active = old.active;
color = old.color;
pc = old.pc;
instr = old.instr;
opcode = instr>>13;
rd = old.rd;
qd = old.qd;
set_reg(rd,qd);
}


void WBXX::show(FILE *out)
{
	if (active) {
		showpc(out,pc,color);
		fprintf(out,"<br>%04x",qd);
	}
	else fprintf(out,"<td><br>");
}

